import java.io.*;
import lejos.nxt.remote.*;
import lejos.pc.comm.*;
import java.util.Arrays;
import java.util.HashSet;
import java.net.URLDecoder;
import lejos.pc.tools.*;
import java.util.Vector;
import java.util.Iterator;


public class NXTMessenger extends Messenger {

	private static final String CONNECTION_STATUS_QUERY = "Connection status?";
	private static final String CONNECTION_STATUS_RESPONSE = "Connection status: ";
	private static final String GENERIC_ERROR = "Generic Error: ";

	private static final String NXT_COMMAND_PREFIX = "NXT Command: ";
	private static final String COMPILER_STATUS_PREFIX = "NXT Compilation Status: ";
	private static final String COMPILER_ERROR_PREFIX = "NXT Compilation Error: ";
	private static final String COMPILER_DONE = "NXT Compilation Finished.";
	private static final String FLASH_STATUS_PREFIX = "NXT FirmwareFlasher Status: ";
	private static final String FLASH_ERROR_PREFIX = "NXT FirmwareFlasher Error: ";
	private static final String FLASH_DONE = "NXT FirmwareFlasher Finished.";
	private static final String NOT_CONNECTED = "NXT is not connected.";

	private static final String FIRMWARE_IS_NOT_LEJOS = "Firmware is wrong.";
	private static final String FIRMWARE_IS_WRONG_LEJOS = "Firmware is wrong version: ";

	private String nxtname = "";
	private String address = "";
	private int protocols = NXTCommFactory.USB; // | NXTCommFactory.BLUETOOTH;
	private NXTCommand nxtcommand;
	private NXTConnector nxtconnector = new NXTConnector();

	private boolean connection_lost = true;
	private boolean uploading = false;
	
	private boolean alerted_user_to_faulty_connection = false;	// true if we already let user know of problem

	static String nxjc;    /** command to run nxjc */
	static String nxjlink; /** command to run nxjlink */

	static {
		if (System.getProperty("os.name").toLowerCase().contains("win")) {
			nxjc = "nxjc.bat";
			nxjlink = "nxjlink.bat";
		} else {
			nxjc = "nxjc";
			nxjlink = "nxjlink";
		}
	}

	public NXTMessenger(String name, Dispatcher dispatcher,
						int connectionDelay, int verbosity) {
		super(name, dispatcher, connectionDelay, verbosity);
	}

	/** Repeatedly tries to connect to an NXT.
	    Does not return unless a connection is established. */
	public void connect() {
		boolean connected = false;
		while (!connected) {
			try {
				sleep(connectionDelay);
			} catch (InterruptedException e) {
				// we are unlikely to be interrupted
				throw new RuntimeException(e);
			}
			//if (verbosity > 8) {
				System.out.println("Trying to connect to NXT");
			//}

			try {
				connected = nxtconnector.connectTo(nxtname, address, protocols);
			} catch (java.lang.RuntimeException e) {
				System.out.println("RuntimeException occurred while trying to connect to NXT: " + e.getMessage());
				if (e.getMessage().equals("Can't initialize bluetooth support")) {
					narrowcastMsg(GENERIC_ERROR + "Bluetooth is turned off on this computer.  Disabling bluetooth scanning.");
					protocols = NXTCommFactory.USB;
				}
			}
		}
		System.out.println("Connected to NXT");
		nxtcommand = new NXTCommand(nxtconnector.getNXTComm()); /// for leJOS 0.9.1
	}

	/** Listens for messages from the NXT,
	    and forwards them to the dispatcher.  Does not exit
	    unless communication is broken. */
	public void receiveMessages() {
		// Right now the NXT does not send any messages to us
		// at this time.  However, unless we know that communication
		// has been lost, this function continues.


		lostConnection(false, "Started receivingMessages()");
		System.out.println("NXT started receiving messages");

		// Check for the right firmware version
		try {
			FirmwareInfo fi = nxtcommand.getFirmwareVersion();
			System.out.println("Firmware version: " + fi.firmwareVersion);
			System.out.println("Firmware protocol: " + fi.protocolVersion);
			System.out.print("Firmware status: " );
			System.out.println(fi.status);
						
			if (fi.status != 0 || fi.firmwareVersion == null || !fi.firmwareVersion.startsWith("1.")) {
				lostConnection(true, "Bad result reading firmware version");
				askUserToRestartRobot();
			} else {
				alerted_user_to_faulty_connection = false;	// connection is good
				String version = nxtcommand.getNXJFirmwareVersion();
				System.out.println("leJOS Firmware version: " + version);
				if (!version.startsWith("0.9.1")) {
					narrowcastMsg(FIRMWARE_IS_WRONG_LEJOS);
				}
			}
		} catch (IOException e) {
			lostConnection(true, "IOException occurred in while checking firmware version: " + e.getMessage());
		} catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("NXT is not running leJOS.");
			narrowcastMsg(FIRMWARE_IS_NOT_LEJOS);
		}
		
		while (!connection_lost) {
			try {
				sleep(500);
			} catch (InterruptedException e) {
				// we are unlikely to be interrupted
				throw new RuntimeException(e);
			}
			if (uploading) continue;
			//if (true) continue;	


			// Broadcast any message received
			try {
				// broadcast message - processes all in queue
				while(!connection_lost && !uploading) {
					byte[] message;
					try {
						message = nxtcommand.messageRead((byte) 0, (byte) 0, true);
					} catch (LCPException e) {
						if (e.getMessage() == "protocol error") {
							break;	// new way of saying, "There is no data here for you"
						} else {
							throw e;	
						}
					}

					if (message.length == 0) break;

					String msgString = new String(message);
					System.out.println("Broadcasting: " + msgString);
					broadcastMsg(msgString);
				}
				// sensor updates - processes all in queue
				while (!connection_lost && !uploading) {
					byte[] message;
					try {
						message = nxtcommand.messageRead((byte) 1, (byte) 0, true);
					} catch (LCPException e) {
						if (e.getMessage() == "protocol error") { 
							break;	// new way of saying, "There is no data here for you"
						} else {
							throw e;	
						}
					}
			
					if (message.length == 0) break;
		
					String msgString = new String(message);
					System.out.println("sensor update: " + msgString);

					//		updateMsg(msgString);
					// Send a standard update message to Scratch
					forwardTo(new ScratchMessage("sensor-update " + msgString), "Scratch Messenger");
					// and a slightly modified one to Enchanting -- each variable name starts with Enc^.
					forwardTo(new ScratchMessage("sensor-update Enc^" + msgString), "Enchanting Messenger");
				}
			
			} catch (LCPException e) {
				lostConnection(true, "LCPException occurred in receiveMessages(): " + e.getMessage());
				e.printStackTrace();
			} catch (IOException e) {
				lostConnection(true, "IOException occurred in receiveMessages(): " + e.getMessage());
				e.printStackTrace();
			} catch (ArrayIndexOutOfBoundsException e) {
				// Our communication is in an unhappy state!
				lostConnection(true, "ArrayIndexOutOfBoundsException occurred in receiveMessages(): " + e.getMessage());
				e.printStackTrace();
				askUserToRestartRobot();
			}			
						
		}
		System.out.println("NXT stopped receiving messages");
		lostConnection(true, "Exited receiveMessages()");
	}

	/** Returns true is the message is sent successfully (and false if there is no connection) */
	public boolean sendMessage(ScratchMessage msg, Messenger source) {

		System.out.println("Received message: " + msg.getName() + " value = " + msg.getValue());

		// Narrowcast messages are just used for IPC (inter-process communication)
		if (msg.getMessageType() == ScratchMessage.NARROWCAST_MSG)
		{
			return actOnIPCCommand(msg);
		}
		else
		{
			// Message for the NXT
			if (connected && !connection_lost && !uploading) {
				// Send variable updates first -- broadcasts may depend upon them
				if (msg.getMessageType() == ScratchMessage.SENSOR_UPDATE_MSG) {
					try {
						String message = msg.getName() + " " + msg.getValue();

						// if the message is from EnchantingFE, it is a variable update
						// send it on port 1
						if (source.name().equals("Enchanting Messenger")) {
							System.out.println("Sending variable update:" + message);
							nxtcommand.messageWrite(message.getBytes(), (byte) 1);
						}
						// if it is from Scratch, it is a remote variable update
						// send it on port 2
						if (source.name().equals("Scratch Messenger")) {
							System.out.println("Sending remote sensor update:" + message);
							nxtcommand.messageWrite(message.getBytes(), (byte) 2);
						}
					} catch (IOException e) {
						lostConnection(true, "IOException while sending sensor update message: " + e.getMessage());
					}
				} else if (msg.getMessageType() == ScratchMessage.BROADCAST_MSG) {
					try {
						String message = msg.getName();
						System.out.println("Sending broadcast:" + message);
						nxtcommand.messageWrite(message.getBytes(), (byte) 0);
					} catch (IOException e) {
						lostConnection(true, "IOException while sending broadcast message: " + e.getMessage());
					}
				}
			}

			return true;
		}
	}

	private boolean actOnIPCCommand(ScratchMessage msg)
	{
		if (msg.getName().startsWith(NXT_COMMAND_PREFIX)) {
			String full_command = msg.getName().substring(NXT_COMMAND_PREFIX.length());
			String[] argv = full_command.split(" ");
			String command = argv[0];	// The command is the first word in the narrowcast message

			System.out.println("Received NXT command: " + Arrays.toString(argv));

			if (command.equals("compile-and-upload")) {
				return compileProgram(argv, false);
			} else if (command.equals("compile-and-run")) {
				return compileProgram(argv, true);
			} else if (command.equals("flash")) {
				return flashFirmware();
			} else if (command.equals("bluetooth-enable")) {
				protocols = NXTCommFactory.USB | NXTCommFactory.BLUETOOTH;
				System.out.println("Scanning using bluetooth enabled.");
			} else if (command.equals("bluetooth-disable")) {
				protocols = NXTCommFactory.USB;
				System.out.println("Scanning using bluetooth disabled.");
			} else if (command.equals("stop-program")) {
				return stopProgram();
			} else if (command.equals("cancel")) {
				return false;  // Front-End sends a signal to quit compiling, but we ignore it for now
			} else {
				System.out.println("Unknown command for NXTMessenger: " + full_command);
			}
		}
		else
		{
			// This isn't an NXT-specific message
			if (msg.getName().equals(CONNECTION_STATUS_QUERY)) {
				sendConnectionStatus();
			}
		}

		return false;
	}

	private void lostConnection(boolean lost, String reason) {

		if (connection_lost == lost) {
			System.out.println("Ignoring lostConnection(" + String.valueOf(connection_lost) + "): " + reason);
			return;
		}

		connection_lost = lost;
		System.out.println("lostConnection(" + String.valueOf(connection_lost) + "): " + reason);
		sendConnectionStatus();

		if (lost == true) {
			try {
				nxtconnector.close();
			} catch (IOException e) {
				System.out.println("IOException while closing connection: " + e.getMessage());
			}
		}
	}
	
	private void askUserToRestartRobot() {
		if (!alerted_user_to_faulty_connection) {
			narrowcastMsg(GENERIC_ERROR + "Faulty communication detected.  Please restart your NXT.");
			alerted_user_to_faulty_connection = true;	
			System.out.println("Asked user to restart NXT");
		}	
	}

	private void amUploading(boolean up, String reason) {
		uploading = up;
		System.out.println("Uploading = " + String.valueOf(uploading) + " - " + reason);
		sendConnectionStatus();
	}

	private void sendConnectionStatus() {
		boolean is_connected = connected && !connection_lost;
		boolean is_ready = is_connected && !uploading;
		narrowcastMsg(CONNECTION_STATUS_RESPONSE + String.valueOf(is_connected) + " " + String.valueOf(is_ready));
	}

	private boolean amConnected() {
		return connected && !connection_lost;
	}

	private boolean compileProgram(String[] argv, boolean run_program) {
		// We are going to compile and upload this code

		// Inform the front end
		amUploading(true, "Starting to upload nxj file");
		narrowcastMsg(COMPILER_STATUS_PREFIX + "Compiling.");

		// Determine the file name.  (It is the only parameter, URL encoded)
		File sourceFile;
		try {
			sourceFile = new File(URLDecoder.decode(argv[1], "UTF-8"));
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException(e);
		}
		if (!sourceFile.exists()) {
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Compilation Failed. " +
				"Source file does not exist. (" + sourceFile.getAbsolutePath() + ")");
			amUploading(false, "Compiling failed -- source file not found.  Abandoning upload.");
			return true;	// acted on message
		}
		System.out.println("File exists? " + sourceFile.exists());

		// Compile the file
		SaferExec se = new SaferExec(300);	// 5 minutes max time to compile
		System.getenv("NXJ_HOME");
		String result = se.execV(nxjc, "-cp", System.getenv("ENCHANTING_JAR"), sourceFile.getPath());
		//System.out.println("Result: *" + result + "*");
		if (!result.equals("0\n")) {
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Compilation Failed. " +
				"Error in code generated by Enchanting.");
			amUploading(false, "Compiling failed.  Abandoning upload.");
			return true;	// acted on message
		}

		// Link the file
		narrowcastMsg(COMPILER_STATUS_PREFIX + "Linking.");
		String name = sourceFile.getName();
		String classname = name.substring(0, name.length() - 5);	// remove ".java"
		File nxjFile = new File(sourceFile.getParent(), classname + ".nxj");
		result = se.execV(nxjlink, "-cp", 
			sourceFile.getParent() + File.pathSeparator + System.getenv("ENCHANTING_JAR"), 
			"-o", nxjFile.getPath(), classname);
		System.out.println("Result: " + result);
		if (!result.equals("0\n")) {
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Linking Failed. ");
			amUploading(false, "Linking failed.  Abandoning upload.");
			return true;	// acted on message
		}

		if (!amConnected()) {
			// wait forty-five seconds to see if we reconnect
			narrowcastMsg(COMPILER_STATUS_PREFIX + "Reconnecting.");

			for (int i = 0; i < 450; ++i) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					throw new RuntimeException(e); 	// we are unlikely to be interrupted
				}
				if (amConnected()) break;
			}

			if (!amConnected()) {
				narrowcastMsg(COMPILER_ERROR_PREFIX + "Unable to reconnect.");
				System.out.println("Unable to reconnect to NXT");
				amUploading(false, "Unable to reconnect.  Abandoning upload.");
				return true;	// acted on message
			}
		}

/*		// wait a little bit longer for connection to be solid
		try {
			sleep(250);
		} catch (InterruptedException e) {
			throw new RuntimeException(e); 	// we are unlikely to be interrupted
		}
*/

		// Upload the file and run it
		if (nxjFile.exists() && amConnected()) {

			String nxtFileName = "Enchanting.nxj";
			try {
				int volume = nxtcommand.getVolume();
				nxtcommand.setVolume((byte)0);
				
				deleteFiles();
				uploadFiles(nxtFileName, sourceFile.getParentFile().getPath() + File.separator);
				nxtcommand.setVolume((byte)volume);
				nxtcommand.playTone(400, 75);	try { sleep(75); } catch (Exception e) {}
				nxtcommand.playTone(600, 75);	try { sleep(75); } catch (Exception e) {}
				nxtcommand.playTone(500, 75);
				
			} catch (IOException e) {
				narrowcastMsg(COMPILER_ERROR_PREFIX + "Uploading Failed. " + e.getMessage());
				amUploading(false, "IOException while changing NXT's volume: " + e.getMessage());
				//lostConnection(false);
			}
	
			try {
				if (run_program) {
					narrowcastMsg(COMPILER_STATUS_PREFIX + "Running.");
					nxtcommand.startProgram(nxtFileName);
					// Connection is lost when programs run
					lostConnection(true, "Need to reconnect now that new application is starting.");
				}
				amUploading(false, "nxj file uploaded successfully");
				narrowcastMsg(COMPILER_DONE);
			} catch (IOException e) {
				narrowcastMsg(COMPILER_ERROR_PREFIX + "Uploading Failed. " + e.getMessage());
				amUploading(false, "IOException while running nxj file: " + e.getMessage());
				//lostConnection(false);
			}
		} else {
			String error_message;
			if (!amConnected()) {
				error_message = "Uploading Failed.  No NXT is connected.";
			} else {
				error_message = "Upload Failed. NXT executable file does not exist. (" + nxjFile.getAbsolutePath() + ")";
			}
			narrowcastMsg(COMPILER_ERROR_PREFIX + error_message);
			System.out.println(error_message);
		}

		return true;	// we acted on the message
	}

	private boolean flashFirmware() {
		amUploading(true, "Flashing firmware");
		FirmwareFlasher f = new FirmwareFlasher();
		try {
			f.flash();
			System.out.printf("Flashed successfully.  Severing connection.");
			lostConnection(true, "Firmware flashed; must reconnect");
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		amUploading(false, "Done flashing firmware");
		return true;
	}

	private boolean stopProgram() {		
		System.out.println("Stopping program");
		try {
			nxtcommand.stopProgram();
			lostConnection(true, "Stopped program; connection lost.");
		} catch (IOException e) {
			lostConnection(true, "IOException while stopping program: " + e.getMessage());
		}
		return true;	// acted on message
	}

	private void uploadFiles(String programFilename, String mediaDirectory) {

		// MEDIA FILE UPLOADING
		narrowcastMsg(COMPILER_STATUS_PREFIX + "Uploading files.");

		String path = mediaDirectory;

		// Upload each individual file
		try {
			HashSet<String> assets = new HashSet<String>();

			// Get the media list that was output by enchanting
			BufferedReader mediaList = new BufferedReader(new FileReader(path + "mediaList.dat"));
			while (mediaList.ready()) {	// While not EOF
				// Each line contains a file name; add it to our set of files.
				assets.add(mediaList.readLine());
			}
			mediaList.close();

			// First, upload the program
			System.out.println("Uploading " + programFilename);
			nxtcommand.uploadFile(new File(path + programFilename), programFilename);
			
			// Then upload all the files from the set
			Iterator itr = assets.iterator();
			nxtcommand.setVerify(true);
			while(itr.hasNext()){
				String filename = (String)itr.next();
				System.out.println("Uploading " + filename);
				File asset = new File(path + filename);
				nxtcommand.uploadFile(asset, filename);
			}
			nxtcommand.setVerify(false);
		} catch (Exception e) {
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Exception occurred processing media: " + e.getMessage());
			System.out.println("Got an IOException in uploadMedia(): " + e.getMessage());
		}
	}
	
	// Deletes all the files off of the NXT
	private boolean deleteFiles() {
	
		narrowcastMsg(COMPILER_STATUS_PREFIX + "Deleting files.");
		HashSet<String> toDelete = new HashSet<String>();	
	
		try {
			FileInfo fi = nxtcommand.findFirst("*.*");
			while (fi != null) {
				if (fi.fileName.startsWith("Enchanting")) 	{ toDelete.add(fi.fileName); }
				if (fi.fileName.startsWith("sprite")) 		{ toDelete.add(fi.fileName); }
				if (fi.fileName.startsWith("stage")) 		{ toDelete.add(fi.fileName); }				
				fi = nxtcommand.findNext(fi.fileHandle);
			}
			
			Iterator itr = toDelete.iterator();
			while(itr.hasNext()){
				String filename = (String)itr.next();
				System.out.println("  Deleting " + filename);
				nxtcommand.delete(filename);
			}
			
			return true;
		} catch (IOException e) {
			narrowcastMsg(COMPILER_ERROR_PREFIX + "Exception occurred deleting files: " + e.getMessage());
			System.out.println("Got an IOException in deleteFiles(): " + e.getMessage());			
			return false;
		}
	}

	private class FirmwareFlasher implements NXJFlashUI {
		NXJFlashUpdate updater = new NXJFlashUpdate(this);

		public void message(String msg) {
			progress(msg, -1);
		}

		public void progress(String msg, int percent) {
			narrowcastMsg(FLASH_STATUS_PREFIX + Integer.toString(percent) + " " + msg);
			//System.out.printf("%s %3d%%\n", msg, percent);
		}

		public void flash() throws Exception {

			boolean simulate = false;

			if (simulate) {
				System.out.println("PRETENDING TO FLASH FIRMWARE");

				// There are many more messages sent when this happens for real
				// but this is a rough approximation
				message("Building firmware image.");
				message("Attempting to reboot the device.");
				message("Locating device in firmware update mode.");
				for (int i = 0; i < 100; ++i) { progress("Searching", i); }
				message("Found NXT: %%NXT-SAMBA%%");
				message("Opened device in firmware update mode.");
				message("Writing firmware image.");
				for (int i = 0; i < 100; ++i) { progress("Writing", i); }
				message("Verifying firmware.");
				for (int i = 0; i < 100; ++i) { progress("Verifying", i); }
				message("Verified 94208 bytes ok.");
				message("Restarting the device.");

			} else {
				System.out.println("FLASHING FIRMWARE");

				File vmFile = null;
				File menuFile = null;
				String home = System.getProperty("nxj.home");
				if (home != null) home = System.getenv("NXJ_HOME");
				byte[] memoryImage = updater.createFirmwareImage(vmFile, menuFile, home);
				byte[] fs = null;
			//	if (commandLine.hasOption("f"))
			//		fs = updater.createFilesystemImage();

				// reset the NXT and open it in firmware upgrade mode
				NXTInfo nxtinfo = nxtconnector.getNXTInfo();
				nxtconnector.close();
				updater.resetDevice(nxtinfo);
				NXTSamba nxt = updater.openSambaDevice(30000);
				if (nxt != null) {
					updater.updateDevice(nxt, memoryImage, fs, true, true, true);
				}
			}
		}
	}
}