Object subclass: #ThumbnailCache
	instanceVariableNames: 'directory fileSizes dict '
	classVariableNames: 'ThumbnailFileName '
	poolDictionaries: ''
	category: 'Scratch-UI-Dialogs'!

!ThumbnailCache methodsFor: 'initialization' stamp: 'jm 1/21/2007 11:32'!
directory: aFileDirectory

	directory _ aFileDirectory.
	dict _ Dictionary new.

	fileSizes _ Dictionary new.
	directory entries do: [:entry |
		entry isDirectory ifFalse: [
			fileSizes at: entry name put: entry fileSize]].
! !


!ThumbnailCache methodsFor: 'other' stamp: 'jm 4/26/2007 18:52'!
filesToUpdate
	"Answer a list of image files that need to have their thumbnails updated."

	| extensions imageFiles |
	extensions _ #(gif jpeg jpg bmp png sprite) asSet.
	imageFiles _ directory fileNames select: [:fn |
		extensions includes: (FileDirectory extensionFor: fn) asLowercase asSymbol].

	^ imageFiles select: [:fn | (dict includesKey: fn) not].
! !

!ThumbnailCache methodsFor: 'other' stamp: 'jm 1/21/2007 11:15'!
readThumbnailFile
	"Attempt to read the thumbnail file in my directory and store the result in dict. If the file doesn't exist or can't be read, make dict be an empty dictionary."

	| fName f buf |
	dict _ Dictionary new.
	fName _ directory fullNameFor: ThumbnailFileName.
	[	(FileStream isAFileNamed: fName) ifTrue: [
			f _ (FileStream readOnlyFileNamed: fName) binary.
			buf _ f contentsOfEntireFile.
			dict _ ObjStream new readObjFrom: (ReadStream on: buf)].
	] ifError: [].
! !

!ThumbnailCache methodsFor: 'other' stamp: 'jm 1/21/2007 11:32'!
removeObsoleteEntries
	"Remove obsolete entries from the dictionary. A thumbnail cache entry is obsolete if (a) the file is no longer found in the directory or (b) the size of the file does not match the files size in the cache entry. Return true if any entries were removed."
	"Note: Originally we used file modification times to detect when a cache entry became obsolete (e.g. because a file was edited). However, these timestamps could change with transitions to/from daylight savings time and when moving laptops between timezones. We now use the file size which is very likely to change when a graphics file is edited. Although there is a small change that an invalid cache entry will not be detected, in that case the obsolete chache entry will probably be very similar in appearance to the edited image. In the worst case, one can manually delete the thumbnails cache file to force the cache to be regenerated."

	| didRemove fSize |
	didRemove _ false.
	dict keys do: [:fn |
		fSize _ fileSizes at: fn ifAbsent: [0].
		((dict at: fn) at: 1) ~= fSize ifTrue: [
			didRemove _ true.
			dict removeKey: fn]].
	^ didRemove
! !

!ThumbnailCache methodsFor: 'other' stamp: 'jm 5/5/2007 18:02'!
scriptCountFor: aSprite
	"Answer the number of scripts in the given sprite. The scripts are in tuple form. A script must start with a hat block."

	| hats count stack |
	hats _ #(EventHatMorph KeyEventHatMorph MouseClickEventHatMorph WhenHatBlockMorph).
	count _ 0.
	aSprite blocksBin do: [:pair |
		stack _ pair at: 2.
		((stack size > 0) and: [stack first size > 0]) ifTrue: [
			(hats includes: stack first first) ifTrue: [count _ count + 1]]].
	^ count

! !

!ThumbnailCache methodsFor: 'other' stamp: 'jm 7/18/2006 12:55'!
thumbnailFor: fileName

	^ (dict at: fileName ifAbsent: [^ nil]) at: 2
! !

!ThumbnailCache methodsFor: 'other' stamp: 'jm 4/26/2007 20:42'!
thumbnailFromImageData: aByteArray
	"Answer a thumbnail for the image stored in the given ByteArray (the contents of a binary file)."

	| f scale |
	f _ Form fromBinaryStream: (ReadStream on: aByteArray).
	f depth < 16 ifTrue: [f _ f asFormOfDepth: 16].
	scale _ (80.0 / f width) min: (60.0 / f height).
	f _ f magnifyBy: scale.
	^ f
! !

!ThumbnailCache methodsFor: 'other' stamp: 'jens 8/14/2009 13:27'!
thumbnailFromSpriteData: aByteArray
	"Answer a thumbnail for the sprite stored in the given ByteArray (the contents of a binary file)."

	| s version proj sprite scriptCount f scale thumbF spriteBlock p label labelForm |
	s _ ReadStream on: aByteArray.

	version _ ObjStream scratchFileVersionFrom: (s next: 10) asString.
	version = 0 ifTrue: [
		ScriptableScratchMorph decodeAsByob.	"make sure we initialize every field - this breaks compatibility with Scratch-sprites. -Jens". 

		s position: 0.
		proj _ ObjStream new readObjFrom: s showProgress: false].
	(version = 1) | (version = 2) ifTrue: [
		s skip: s uint32.  "skip header"
		proj _ ObjStream new readObjFrom: s showProgress: false].

	proj class = ScratchStageMorph ifFalse: [^ nil].
	sprite _ proj submorphs detect: [:m | m isKindOf: ScriptableScratchMorph] ifNone: [^ nil].
	f _ sprite costumeForm asFormOfDepth: 16.
	scale _ (56.0 / f width) min: (42.0 / f height).
	f _ f magnifyBy: scale.
	thumbF _ Form extent: 80@60 depth: 16.
	f displayOn: thumbF at: ((thumbF width - f width) // 2) @ 0 rule: Form paint.
	spriteBlock _ ScratchFrameMorph skinAt: #spriteBlock.
	p _ ((thumbF width - spriteBlock width) // 2) @ (thumbF height - spriteBlock height).
	spriteBlock displayOn: thumbF at: p rule: Form paint.

	scriptCount _ self scriptCountFor: sprite.
	label _ 'Scripts' localized, ScratchTranslator colonSuffix, ' ', scriptCount printString.
	scriptCount = 0 ifTrue: [label _ 'no scripts' localized].
	scriptCount = 1 ifTrue: [label _ '1 script' localized].
	labelForm _ (StringMorph contents: label font: (StrikeFont fontName: 'Verdana' size: 10)) imageForm.
	labelForm displayOn: thumbF at: p + (((spriteBlock width - labelForm width // 2) + 1)@ 5) rule: Form paint.

	^ thumbF
! !

!ThumbnailCache methodsFor: 'other' stamp: 'CWB 7/23/2011 12:31'!
updateThumbnails
	"Update my dictionary by creating thumbnails for all the image files in my directory. If updates were needed, try to save save the thumbnails cache."

	| didRemove fList n thumb buf fSize |
	didRemove _ self removeObsoleteEntries.
	fList _ self filesToUpdate.

	fList size > 0 ifTrue: [
		n _ 0.
		('Updating thumbnails' localized, ScratchTranslator ellipsesSuffix)
			displayProgressAt: Display center
			from: 0 to: fList size
			during: [:bar |
				fList do: [:fn |
					bar value: n.
					thumb _ nil.
					[
						buf _ (FileStream readOnlyFileNamed: (directory fullNameFor: fn)) binary contentsOfEntireFile.
						(fn asLowercase endsWith: '.sprite')
							ifTrue: [thumb _ self thumbnailFromSpriteData: buf]
							ifFalse: [thumb _ self thumbnailFromImageData: buf].
					] ifError: [].
					fSize _ fileSizes at: fn ifAbsent: [0].
					thumb ifNotNil: [dict at: fn put: (Array with: fSize with: thumb colorReduced)].
					n _ n + 1]]].

	(didRemove | (fList size > 0) | (dict size = 0)) ifTrue: [self writeThumbnailFile].
! !

!ThumbnailCache methodsFor: 'other' stamp: 'jm 11/7/2006 11:08'!
writeThumbnailFile
	"Write dict to the thumbnail file in my directory. Do nothing if the file cannot be written."

	| f |
	dict size = 0 ifTrue: [  "if no thumbnails and no thumbnails file to delete, do nothing"
		(FileStream isAFileNamed: (directory fullNameFor: ThumbnailFileName))
			ifFalse: [^ self]].

	[	directory deleteFileNamed: ThumbnailFileName.  "delete may fail if directory is read-only or is an alias"
		((dict size > 0) or: [directory fileExists: ThumbnailFileName]) ifTrue: [
			f _ (FileStream fileNamed: (directory fullNameFor: ThumbnailFileName)) binary.
			ObjStream new storeObj: dict on: f.
			f close].
	] ifError: [].
! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ThumbnailCache class
	instanceVariableNames: ''!

!ThumbnailCache class methodsFor: 'class initialization' stamp: 'jm 7/16/2006 19:56'!
initialize
	"self initialize"

	ThumbnailFileName _ 'scratchthumbs.db'.
! !


ThumbnailCache initialize!
